""" types """

from datetime import datetime
from enum import Enum
from locale import currency
import locale
from typing import List, Optional, TypedDict, Union
from pydantic import BaseModel

Content = Union[bytes, str]


class Account:
    phone: str
    internet: str

    def __str__(self) -> str:
        return f"DSL# {self.internet} | Phone# {self.phone}"

    def __repr__(self) -> str:
        return self.__str__()


class BillStatus(Enum):
    UNKNOWN = 0
    PAID = 1
    UNPAID = 2


class BillAmount:
    amount: float
    currency: str

    def __init__(
        self, amount: Optional[float] = 0, currency: Optional[str] = "LBP"
    ) -> None:

        self.amount = amount
        self.currency = currency

        if currency in ["L.L.", "LL", "L.L"]:
            self.currency = "LBP"

    @staticmethod
    def parse(str_val: str):
        amount, currency = str_val.split(" ")
        amount = float(amount.replace(",", ""))

        return BillAmount(amount, currency)

    def __str__(self) -> str:
        return f"{str(self.amount)} {self.currency}"

    def __repr__(self) -> str:
        return self.__str__()


class Bill:
    date: datetime
    amount: BillAmount
    status: BillStatus

    def __str__(self) -> str:
        if self.status == BillStatus.PAID:
            status = "paid"
        elif self.status == BillStatus.UNPAID:
            status = "not paid"
        else:
            status = "unknown"

        return f"Bill [{self.date.strftime('%b %Y')}], {self.amount}: {status}"

    def __repr__(self) -> str:
        return self.__str__()


class BillInfo:
    total_outstanding: BillAmount
    bills: List[Bill] = []

    def __str__(self) -> str:
        return f"Total outstanding: {self.total_outstanding}"

    def __repr__(self) -> str:
        return self.__str__()


class ConsumptionInfo:
    speed: str
    quota: int
    upload: float
    download: float
    total_consumption: float
    extra_consumption: float
    last_update: datetime

    def __str__(self) -> str:
        return f"Total Consumption: {self.total_consumption} GB; Last update: {self.last_update}"

    def __repr__(self) -> str:
        return self.__str__()


class ErrorResponseContent(TypedDict):
    code: Union[int, str]


class ErrorResponse(TypedDict):
    error: ErrorResponseContent


class LoginResponse(TypedDict):
    SessionID: str


class ConfigUser(BaseModel):
    username: str
    password: str


class OgeroConfigFile(BaseModel):
    """config file definition"""

    users: List[ConfigUser]



""" constants """
from pathlib import Path
from typing import TypedDict
from zoneinfo import ZoneInfo

BASEURL = {
    "api": "https://ogero.gov.lb/API",
    "myogero": "https://ogero.gov.lb/myogero",
}

CERT_PATH = Path(__package__, 'assets/ogero.pem')


DefaultHeaders = TypedDict(
    "DefaultHeaders",
    {
        "Accept": str,
        "Cache-Control": str,
        "Content-Type": str,
        "Origin": str,
        "Referer": str,
    },
)


def default_headers() -> DefaultHeaders:
    """returns a default set of headers"""
    return {
        "Accept": "application/json",
        # "Content-Type": "application/json",
        # "Origin": "https://ogero.gov.lb",
        # "Referer": "https://ogero.gov.lb/",
        "Cache-Control": "no-cache",
    }


API_ENDPOINTS = {
    "login": f'{BASEURL["api"]}/Login.php',
    "dashboard": f'{BASEURL["myogero"]}/mobileapp.dashboard.php?SessionID={{session_id}}&Username={{username}}&AppRequest&nbr={{phone_account}}&dsl={{internet_account}}',
    "bill": f'{BASEURL["myogero"]}/bill.php?SessionID={{session_id}}&Username={{username}}&AppRequest&nbr={{phone_account}}',
    "consumption": f'{BASEURL["myogero"]}/consumption.php?SessionID={{session_id}}&Username={{username}}&AppRequest&dsl={{internet_account}}',
}

CONNECTION_SPEED = "Connection Speed"
QUOTA = "Quota"
UPLOAD = "Upload"
DOWNLOAD = "Download"
TOTAL_CONSUMPTION = "Total Consumption"
EXTRA_CONSUMPTION = "Extra Consumption"
LAST_UPDATE = "Consumption Until"

LEBANON_TIMEZONE = ZoneInfo('Asia/Beirut')


""" shared utilities """
from datetime import datetime
import logging
from typing import List
from bs4 import BeautifulSoup, ResultSet, Tag
from .const import (
    CONNECTION_SPEED,
    DOWNLOAD,
    EXTRA_CONSUMPTION,
    LAST_UPDATE,
    QUOTA,
    TOTAL_CONSUMPTION,
    UPLOAD,
    LEBANON_TIMEZONE
)
from .types import (
    Account,
    Bill,
    BillAmount,
    BillInfo,
    BillStatus,
    Content,
    ConsumptionInfo,
)

def __parse_status_value(str_val: str):
    val = str_val.split()[0]

    if val == "Unlimited":
        return float("inf")
    else:
        return float(val)

def __parse_bill_status(amount_tag: Tag, bill: Bill) -> None:
    col_status = amount_tag.parent.findNextSibling("td")
    status = col_status.get_text(strip=True).lower()
    if status == "paid":
        bill.status = BillStatus.PAID
    elif status == "not paid":
        bill.status = BillStatus.UNPAID
    else:
        bill.status = BillStatus.UNKNOWN
    logging.debug(f"status: {bill.status}")


def __parse_bill(bill_tag: Tag):
    bill = Bill()
    col_date = bill_tag.find(class_="BillDate")
    if col_date is None:
        return None

    date_text = col_date.get_text(strip=True)
    bill.date = datetime.strptime(date_text, "%b%Y")

    logging.debug(f"date: {bill.date.strftime('%b %Y')}")

    col_amount = bill_tag.find(class_="BillAmount")
    amount_str = col_amount.get_text(strip=True)
    bill.amount = BillAmount.parse(amount_str)
    logging.debug(f"amount: {bill.amount}")

    __parse_bill_status(col_amount, bill)

    return bill


def parse_content(content: Content):
    """
    convert html into soup
    """
    return BeautifulSoup(content, "html.parser")


def parse_consumption_info(content: Content) -> ConsumptionInfo:

    info = ConsumptionInfo()
    statuses = parse_content(content).find_all(class_="MyConsumptionGrid")

    for status_div in statuses:
        [key, value] = [span.text.strip() for span in status_div]

        if key == CONNECTION_SPEED:
            info.speed = value
        elif key == QUOTA:
            info.quota = __parse_status_value(value)
        elif key == UPLOAD:
            info.upload = __parse_status_value(value)
        elif key == DOWNLOAD:
            info.download = __parse_status_value(value)
        elif key == TOTAL_CONSUMPTION:
            info.total_consumption = __parse_status_value(value)
        elif key == EXTRA_CONSUMPTION:
            info.extra_consumption = __parse_status_value(value)
        elif key == LAST_UPDATE:
            info.last_update = datetime.strptime(value, "%B %d,%Y  %H:%M").replace(tzinfo=LEBANON_TIMEZONE)

    return info


def parse_accounts(content: Content) -> List[Account]:

    accounts: List[Account] = []

    account_options = (
        parse_content(content).find("select", id="changnumber").find_all("option")
    )

    for account_option in account_options:

        account = Account()
        account.phone = account_option.attrs["value"]
        account.internet = account_option.attrs["value2"]

        accounts.append(account)

    return accounts


def parse_bills(content: Content) -> BillInfo:

    bill_info = BillInfo()

    bill_outstanding_section = parse_content(content).find(
        class_="BillOutstandingSection1"
    )

    if bill_outstanding_section is not None:
        bill_outstanding_val =  bill_outstanding_section.find("span").get_text(
            strip=True
        )

        bill_info.total_outstanding = BillAmount.parse(bill_outstanding_val)
    else:
        bill_info.total_outstanding = BillAmount()

    logging.debug(f"outstanding {bill_info.total_outstanding}")

    bill_table = parse_content(content).find("table", class_="BillTable")
    bill_rows: List[Tag] = bill_table.find_all("tr")

    for row in bill_rows:

        logging.debug(f"################################")
        bill = __parse_bill(row)

        if bill is None:
            continue

        bill_info.bills.append(bill)

    logging.debug(f"################################")

    return bill_info


def parse_error_message(content: Content):

    script_tag = parse_content(content).find("script", {"language": "javascript"})

    if script_tag is None:
        return None

    msg = script_tag.get_text(strip=True)

    err_idx = msg.find("error=")
    if err_idx == -1:
        return None

    err_msg = msg[err_idx + 6 :]
    err_msg = err_msg.split("&")[0]
    err_msg = err_msg.split(";")[0]
    err_msg = err_msg.split('"')[0]

    return err_msg



""" aiohttp support for Ogero """

import logging
import sys
from typing import List, Optional
import ssl

from ..types import Account, BillInfo, ConsumptionInfo, ErrorResponse, LoginResponse
from ..utils import (
    parse_accounts,
    parse_bills,
    parse_consumption_info,
    parse_error_message,
)
from ..const import API_ENDPOINTS, CERT_PATH, DefaultHeaders, default_headers
from ..exceptions import AuthenticationException

try:
    import aiohttp
    from aiohttp.client import ClientSession, ClientResponse, TCPConnector
except ImportError as error_message:
    print(f"Failed to import aiohttp, bailing: {error_message}", file=sys.stderr)
    sys.exit(1)


class Ogero:
    """aiohttp class for interacting with Ogero APIs"""

    def __init__(
        self,
        username: str,
        password: str,
        session: Optional[ClientSession] = None,
        debug: bool = False,
        logger: logging.Logger = logging.getLogger(),
    ):
        """Setup function
        ```
        @param username: str - username for Ogero account
        @param password: str - password for Ogero account
        @param debug: bool - debug mode
        @param logger: Logger
        ```
        """

        if not (username and password):
            raise AuthenticationException(
                "You need to supply both username and password"
            )

        self.username = username
        self.password = password
        self.debug = debug
        self.logger = logger
        self.session_id = None

        if not session:
            self.session = ClientSession()
        else:
            self.session = session

        self.ssl_context = ssl.create_default_context(cafile=CERT_PATH)


    async def login(self):
        """Logs into the account and caches the session id."""

        url = API_ENDPOINTS["login"]

        headers = default_headers()
        payload = {"Username": self.username, "Password": self.password}

        async with self.session.post(url, headers=headers, data=payload, ssl=self.ssl_context) as response:
            self.session_id = None
            await self.handle_response_fail(response)
            jsondata: LoginResponse = await response.json()

        self.logger.debug("Login response status: %s", response.status)
        # self.logger.debug("Dumping login response: %s", json.dumps(jsondata))

        self.session_id = jsondata["SessionID"]

        return True

    async def get_accounts(self, account: Account = None) -> List[Account]:
        """Get user phone/internet accounts"""

        url = API_ENDPOINTS["dashboard"]

        response = await self.request_get(url, account)
        await self.handle_response_fail(response)
        content = await response.text()
        accounts = parse_accounts(content)

        self.logger.debug("Dashboard response status: %s", response.status)
        self.logger.debug("Dumping accounts response: %s", accounts)

        return accounts

    async def get_bill_info(self, account: Account = None) -> BillInfo:
        """Get bill info for phone account
        ```
        @param account: Account - Phone/Internet account
        ```
        """

        url = API_ENDPOINTS["bill"]

        response = await self.request_get(url, account)
        await self.handle_response_fail(response)
        content = await response.text()
        bill_info = parse_bills(content)

        self.logger.debug("Bill response status: %s", response.status)
        self.logger.debug("Dumping bill response: %s \n%s", bill_info, bill_info.bills)

        return bill_info

    async def get_consumption_info(self, account: Account = None) -> ConsumptionInfo:
        """Get consumption info for internet account
        ```
        @param account: Account - Phone/Internet account
        ```
        """

        url = API_ENDPOINTS["consumption"]

        response = await self.request_get(url, account)
        await self.handle_response_fail(response)
        content = await response.text()
        consumption_info = parse_consumption_info(content)

        self.logger.debug("Consumption response status: %s", response.status)
        self.logger.debug("Dumping consumption response: %s", consumption_info)

        return consumption_info

    async def request_get(
        self,
        url: str,
        account: Account = None,
        headers: DefaultHeaders = default_headers(),
        max_retries: int = 1
    ):
        """Send get request and check if session is active
        ```
        @param url: str - Endpoint url
        @param account: Account - Phone/Internet account
        @param headers: DefaultHeaders
        ```
        """
        if max_retries < 0:
            return None
        
        if self.session_id is None:
            await self.login()

        try:
            params = self._get_params(account)
            formatted_url = url.format_map(params)
            response = await self.session.get(formatted_url, headers=headers, ssl=self.ssl_context)
            await self.handle_response_fail(response)
        except AuthenticationException as ex:
            self.logger.debug(f"AuthenticationException: {ex}")
            await self.login()
            response = await self.request_get(url, account, headers, max_retries - 1)

        return response

    def _get_params(self, account: Account = None):

        if self.session_id is None:
            raise AuthenticationException("Login first")

        return {
            "session_id": self.session_id,
            "username": self.username,
            "phone_account": account.phone if account else "",
            "internet_account": account.internet if account else "",
        }

    async def handle_response_fail(
        self,
        response: ClientResponse,
    ) -> None:
        """Handles response status codes.
        ```
        @param response - aiohttp.Response - the full response object
        ```
        """

        if response.status == 400 and response.content_type == "application/json":
            resp: ErrorResponse = await response.json()
            msg = resp["error"]["message"]
            self.logger.debug(f"AuthenticationException: {msg}")
            raise AuthenticationException(msg)

        if response.status == 200 and response.content_type == "text/html":
            content = await response.text()
            msg = parse_error_message(content)
            if msg is not None and msg.startswith("You are required to login"):
                raise AuthenticationException(msg)

        response.raise_for_status()
